<?php
/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */
/*
This code is a PHP script that can be used as a chatbot transfer target. It asks the user a series 
of pre-configured questions and then transfers the chat to a human agent for further assistance.

Here's a breakdown of the code:

Configuration:

$DEBUG: Controls the level of debug logging (0 for no logging, 1 for basic logging).

$config: Defines various messages and settings for the chatbot, including:
    last_action:        The action to take after completing all questions (default: transfer).
    invalid_answer_message: Message to display when the user provides an invalid answer.
    close_message:          Message to display when the chatbot closes the conversation.
    transfer_message:       Message to display when the chatbot transfers the chat to a human agent.

$questions: Array of questions to ask the user, each with the following properties:
    body:          The text of the question.
    response_type: The type of answer expected (text, range, or menu).
    options:       Additional information depending on the response type (e.g., valid range for range questions, options for menu questions).
    next_action:   The next step after the user answers the question (e.g., index of the next question, special action like transfer).

Processing Flow:

Retrieve Input:

The script reads the JSON data sent from the chatbot platform.

It retrieves the state field, which indicates the current stage of the conversation.

Validate Answer (if applicable):

If the state indicates a previous question, the answer is validated based on the question type and expected response.
If the answer is invalid, an error message is displayed.
If the answer is valid, the next step is determined.

Retrieve Next Question:
Based on the current state or validated answer, the next question is retrieved from the $questions array.

Respond to User:
The script sends a response to the chatbot platform containing the next question or a message based on the current stage (e.g., close message, transfer message).

Update State:
The state is updated based on the current stage and next step.

Additional Notes:

The script supports different types of questions:
Text: Open-ended questions where the user can type any response.
Range: Questions where the user selects a number from a defined range.
Menu: Questions where the user chooses an option from a list.
The script can be used to collect information from the user before transferring the chat to a human agent.
This is just a basic example, and the script can be customized further to meet specific needs.

*/

require_once("../chatbroker.class.".PHP_MAJOR_VERSION.".php");

$DEBUG=1;

if(!is_callable('array_key_first')) {
   function array_key_first($myarray) {
       reset($myarray);
       return key($myarray);
   }
}

$config = array(
   'max_interactions'       => 5,
   'last_action'            => 'transfer',
   'invalid_answer_message' => 'No entendí la respuesta.',
   'close_message'          => 'Gracias!',
   'transfer_message'       => 'Gracias, lo contactaré con un asistente humano',
   'breakout_word'         => '',
   'breakout_step'         => 1

);

// override all hardcoded options if we have a bot name and ini file
if(isset($_REQUEST['bot'])) {
    $filename = getcwd()."/".$_REQUEST['bot'].".ini";
    if(is_readable($filename)) {

        $ini = parse_ini_file($filename,false);
        $questions = json_decode(base64_decode($ini['steps']),true);

        if(isset($ini['max_interactions'])) {
            if($ini['max_interactions']!='') {
                $config['max_interactions']=intval($ini['max_interactions']);
            }
        }

        if(isset($ini['last_action'])) {
            if($ini['last_action']!='') {
                $config['last_action']=trim($ini['last_action']);
            }
        }

        if(isset($ini['invalid_answer_message'])) {
            if($ini['invalid_answer_message']!='') {
                $config['invalid_answer_message']=trim($ini['invalid_answer_message']);
            }
        }

        if(isset($ini['close_message'])) {
            if($ini['close_message']!='') {
                $config['close_message']=trim($ini['close_message']);
            }
        }

        if(isset($ini['transfer_message'])) {
            if($ini['transfer_message']!='') {
                $config['transfer_message']=trim($ini['transfer_message']);
            }
        }

        if(isset($ini['breakout_word'])) {
            if($ini['breakout_word']!='') {
                $config['breakout_word']=trim($ini['breakout_word']);
            }
        }

        if(isset($ini['breakout_step'])) {
            if($ini['breakout_step']!='') {
                $config['breakout_step']=trim($ini['breakout_step']);
            }
        }

    }
}

/*

$questions = array(
    array(
        "body"          => "Cuál es su nombre?",
        "response_type" => "text",
        "next_action"   => "next"
    ),
    array(
        "body"          => "Del 1 al 5, cómo nos calificaría, siendo 1 muy malo y 5 muy bueno?",
        "response_type" => "range",
        "options"       => array(1,2,3,4,5),
        "next_action"   => "next"
    ),
    array(
        "body"          => "Qué día está libre?",
        "response_type" => "text",
        "next_action"   => "next"
    ),
    array(
        "body"          => "Por favor elija una opción:\n1) Consulta General\n2) Retiro de puntos\n3) Vacuna o Inyectable",
        "response_type" => "menu",
        "options"       => array("1" => 4, "2" => 5, "3" => 6)
    ),
    array(
        "body"          => "Muy bien, ahora elija la especialidad:\nA) Consulta general\nB) Consulta Ortopédica\nC) Consulta de control",
        "response_type" => "menu",
        "options"       => array("A" => 7, "B" => 7, "C" => 7)
    ),
    array(
        "body"          => "Muy bien, que tipo?:\nA) Retiro de puntos de Cirugía Ortopédica\nB) Retiro de Puntos de otra cirugía",
        "response_type" => "menu",
        "options"       => array("A" => 7, "B" => 7)
    ),
    array(
        "body"          => "Muy bien, que tipo de vacuna?:\nA) Vacunas anuales\nB) Vacunas de cachorros\nC) Inyectable de control",
        "response_type" => "menu",
        "options"       => array("A" => 7, "B" => 7, "C" => 7)
    ),
    array(
        "body"          => "Gracias! Lo contactaremos con un asesor de soporte técnico",
        "response_type" => "transfer",
        "options"       => "Soporte",
        "next_action"   => "next"
    )
);

*/

$json = file_get_contents('php://input');
$data = json_decode($json,1);

if(!isset($data)) { exit; }

log_debug("---------------------------------------\ninteractive bot start ".$_REQUEST['bot']);

$validate_command = '';

$interaction_counter = 0;

$answer = trim($data['body']);

if($data['state']=='') {

    // First interation with bot, retrieve first question (index 0 of questions array)
    $step=0;
    log_debug("first interaction as we have no state");
    $botdata=array();
    $botdata['body']=$answer;

} else {

    log_debug("its not the first interaction");
    // Not first interaction, validate answer from previous step
    // the state field will hold the string "previous_step-step"
    // to know where we came from (to validate the answer from a previous question), 
    // and when we are to print the next message/question after validating

    list($prev_step,$step,$botdata,$interaction_counter, $current_wait) = get_state($data['state']);
    $botdata['body']=$answer;

    if($interaction_counter > $config['max_interactions']) {
        if (strpos($config['last_action'], 'transfer') !== false) {
            $return = array("body"=>$config['transfer_message']." [".$config['last_action']."]","attach"=>'',"state"=>'');
        } else if($config['last_action']=='close') {
            $return = array("body"=>$config['close_message']." [close]","attach"=>'',"state"=>'');
        }
        echo json_encode($return);
        die();
    }

    $step      = intval($step);
    $prev_step = intval($prev_step);
    $interaction_counter = intval($interaction_counter);

    log_debug(print_r($data,1));

    log_debug("current step: $step, previous step: $prev_step, counter: $interaction_counter (".$data['state'].")");

    $validate_command = validate_answer($prev_step,$answer);

    log_debug("validate answer ( $prev_step, $answer) returned comand: $validate_command");
}

if($validate_command=='[wrong]') {
    log_debug("wrong");
    list($msg,$state) = retrieve_question($prev_step);
    if(isset($questions[$prev_step]['incorrectmsg']) && $questions[$prev_step]['incorrectmsg']!='') {
        $return = array("body"=>array($questions[$prev_step]['incorrectmsg']),"attach"=>'',"state"=>$state);
    } else {
        $return = array("body"=>array($config['invalid_answer_message'],$msg),"attach"=>'',"state"=>$state);
    }
    echo json_encode($return);
    die();
} else if($validate_command=='[close]') {
    $return = array("body"=>$config['close_message']." [close]","attach"=>'',"state"=>'');
    echo json_encode($return);
    die();
} else if($validate_command=='[transfer]') {
    $return = array("body"=>$config['transfer_message']." [transfer]","attach"=>'',"state"=>'');
    echo json_encode($return);
    die();
} else if(preg_match("/\[next=(\d+)\]/", $validate_command, $matches)) {
    $step = $matches[1];
    log_debug("next step from menu ".$step);
}


// api calls, file attachs or tag calls
$msg_attach = array();
$i=0;
while(true) {
    $i++;
    if($questions[$step]['response_type']=='apicall') {
        log_debug("api call step $step, $i");
        list($step,$apidata) = api_call($step);
        $botdata['apiresult']=$apidata;
    } else if($questions[$step]['response_type']=='file') {
        log_debug("send file step $step, $i");
        list($step,$attach) = attach_file($step);
        $msg_attach[]=$attach;
    } else if($questions[$step]['response_type']=='tag') {
        $step = tag_call($step);
    } else {
        log_debug("break loop api/file on step $i, next step $step");
        break;
    } 

    if(!isset($questions[$step])) { log_debug("break loop as there is no step $step"); break; }

    if( $questions[$step]['response_type']!='apicall' && 
        $questions[$step]['response_type']!='tag' && 
        $questions[$step]['response_type']!='file') { log_debug("break loop as the next type is ".$questions[$step]['response_type']); break; }

    if($i>100) { log_debug("break loop as there are too man¥ attachs or files"); break; }
}

// check route step based on user message
$step_type='';
$i=0;
do {
    $step = follow_route($step,$data['body'],$data['business_hours'],$data['number'],$data['chatid'],$data['from']);
    $step_type = $questions[$step]['response_type'];
    log_debug("return from route, goto step $step, is type $step_type");
    $i++;
} while($step_type=='route' && $i<90);

log_debug("final step after route $step");

list($msg,$state) = retrieve_question($step);

log_debug("retrieve question $step returned  $msg, set state to $state, current answer ($answer)");

// check for answer to wait step
if($questions[$step]['response_type']=='wait') {
    if($step==$prev_step && $interaction_counter>1) {
        // inside a wait step, user wrote before waiting
        $msg = expand_variables($questions[$step]['response']);
        log_debug("received user message while on wait step, respond with $msg");

        // restore wait to original wait question value, so additional responses do not restart the timer, set wait to -1 to NOT update the wait timestamp and alter the wait counter
        $temp_state = json_decode($state,1);
        $temp_state['wait']=-1;
        $state = json_encode($temp_state);
    }
}

// Append configuration close or transfer messages if the command was received
if($msg=='[close]')    $msg = $config['close_message']." [close]";
if($msg=='[transfer]') $msg = $config['transfer_message']." [transfer]";

if(count($msg_attach)>0) {
    log_debug("Has attachments");
    $save_msg = $msg;
    $msg=array();
    foreach($msg_attach as $data) {
        $msg[]='';
    }
    // insert last step that is not a file attach
    $msg[]=$save_msg;
    $msg_attach[]=array();
} 

//log_debug(print_r($msg,1));
//log_debug("state = $state");

// Responds to chatbot with actual message and exists
$return = array("body"=>$msg,"attach"=>$msg_attach,"state"=>$state);
//log_debug(print_r($return,1));
echo json_encode($return);
exit;

function attach_file($step) {
    global $questions, $config, $data;
    if(!isset($questions[$step])) return array("[".$config['last_action']."]","end");
    if($questions[$step]['response_type']=='file') {
        $fileDetails = downloadFile($questions[$step]['body']);
        $step++;
        return array($step,$fileDetails);
    } else {
        return array($step,array());
    }
}

function tag_call($step) {
    global $questions, $config, $data;
    if(!isset($questions[$step])) return array("[".$config['last_action']."]","end");
    if($questions[$step]['response_type']=='tag') {
        $settag = expand_variables($questions[$step]['body']);
        // perform tag call, jump to next step
        $cb = new ChatBroker(array());
        $res = $cb->tagConversation($data['conversationid'],$settag);
        $step++;
    }
    return $step;
}

function get_state($state) {
    $state = json_decode($state);
    $astate = json_decode($state,1);
    $previous_step = $astate['previous_step'];
    $current_step  = $astate['current_step'];
    $interaction_counter  = $astate['interaction_counter'];
    $wait = isset($astate['wait'])?$astate['wait']:0;
    $botdata = $astate['data'];
    return array($previous_step,$current_step,$botdata,$interaction_counter,$wait);
}

function api_call($step) {
    global $questions, $config, $data, $botdata;
    if(!isset($questions[$step])) return array("[".$config['last_action']."]","end");
    if($questions[$step]['response_type']=='apicall') {
        $method  = $questions[$step]['httpmethod'];
        $url     = expand_variables($questions[$step]['url']);
        $body    = expand_variables($questions[$step]['body']);
        $headers = expand_variables($questions[$step]['headers']);
        $goto    = $step+1;
        $array_response = send_request($method, $url, $body, $headers, array());
        //log_debug("request response:\n".print_r($array_response,1));
        return array($goto,json_encode($array_response));
    } else {
        return array($step,"");
    }
}

function follow_route($step,$answer,$business_hours,$number,$chatid,$from) {
    global $questions, $config, $data, $botdata;

    if(!isset($questions[$step])) return array("[".$config['last_action']."]","end");

    $goto = $step;

    if($questions[$step]['response_type']=='route') {

        if(isset($questions[$step]['from']) && isset($questions[$step]['from_goto'])) {
            $textoArray = explode(',',$questions[$step]['from']);
            foreach($textoArray as $texto) {
                if (strpos($chatid, $texto) !== false || strpos($from, $texto) !== false || strpos($number, $texto) !== false) {
                    $goto = $questions[$step]['from_goto'];
                    log_debug("Follow route on from, goto ($goto)");
                    return $goto;
                }
            }
        }

        if(isset($questions[$step]['afterhours'])) {
            if($questions[$step]['afterhours']!='' && $business_hours=='closed') {
                $goto = $questions[$step]['afterhours'];
                log_debug("Follow route out of business hours, goto ($goto)");
                return $goto;
            }
        }

        // check api responses
        if($questions[$step]['apivariable']!='') {
            $apiresult = json_decode($botdata['apiresult'],1);
            if($apiresult!==null) {
                if(isset($apiresult[$questions[$step]['apivariable']])) {
                    if (stripos($apiresult[$questions[$step]['apivariable']],$questions[$step]['apivariablecontent']) !== false) {
                        $goto = $questions[$step]['apivariablegoto'];
                        log_debug("Goto api result $goto");
                        return $goto;
                    }
                } else {
                   log_debug('There is no '.$questions[$step]['apivariable'].' in api response');
                   log_debug(print_r($apiresult,1));
                }
            }
        }

        $valid_answers = array();
        $goto = ($questions[$step]['nomatch']!='')?$questions[$step]['nomatch']:$step+1;
        log_debug("Goto if not match to $goto");
        foreach($questions[$step]['options'] as $key=>$val) {
            $valid_answers[strtolower($key)]=$val;
        }
        $found=0;
        foreach($valid_answers as $key=>$dest) {
            if(preg_match("/$key/i",$answer)) {
                 $found=1;
                 $goto = $dest;
                 break;
            }
        }
        log_debug("Follow route found ($found), goto ($goto)");

   }

    return $goto;
}

function retrieve_question($step) {

    global $questions, $config, $data, $botdata, $interaction_counter;

    $interaction_counter++;

    $attach = array();

    if(!isset($questions[$step])) return array("[".$config['last_action']."]","end");

    $question_body = expand_variables($questions[$step]['body']);

    $wait = 0;
    if($questions[$step]['response_type']=='transfer') {
       if($questions[$step]['options']!='') {
           $question_body.="[transfer^".$questions[$step]['options']."]";
       } else {
           $question_body.="[transfer]";
       }
    } else if($questions[$step]['response_type']=='wait') {
        $wait = intval($questions[$step]['wait']); 
    } else if($questions[$step]['response_type']=='close') {
        $question_body.="[close]";
    } else if($questions[$step]['response_type']=='close^nocsat') {
        $question_body.="[close^nocsat]";
    } else if($questions[$step]['response_type']=='menu') {
        if(strlen(array_key_first($questions[$step]['options']))>1 && strpos($data['chatid'], "meta_") === 0) {
            // convert menu to whatsapp interactive buttons
            // menu option is not a single character/digit, asume interactive message (up to three buttons only)

            if(count($questions[$step]['options'])<4) {
                $buttons = array();
                $i=0;
                foreach($questions[$step]['options'] as $key=>$goto) {
                    if($i>2) break;
                    $buttons[] = array("type"=>"reply","reply"=>array("id"=>"id".$i,"title"=>$key));
                    $i++;
                }

                if($i>0) {
                    $body = array(
                        "type"   =>"button",
                        "body"   => array( "text" => $questions[$step]['body'] ),
                        "action" => array(
                          "buttons" => $buttons
                        )
                    );
                    $unicodeChar = '\u0008';
                    $null = json_decode('"'.$unicodeChar.'"');
                    $question_body=$null.json_encode($body);
                }
            } else {
                if(strpos($data['chatid'], "meta_whatsapp") === 0) {
                    // interactive list type
                    $items = array();
                    $i=0;
                    foreach($questions[$step]['options'] as $key=>$goto) {
                        $items[]=array("id"=>$i,"title"=>$key);
                        $i++;
                    }
                    $body = array(
                        "type" => "list",
                        "body" => array(
                            "text" => $questions[$step]['body']
                            ),
                        "action" => array(
                            "button" => "Options",
                            "sections" => array(
                                array(
                                    "title" => "Section",
                                    "rows" => $items,
                                )
                            )
                        )
                    );
                    $unicodeChar = '\u0008';
                    $null = json_decode('"'.$unicodeChar.'"');
                    $question_body=$null.json_encode($body);
                }
            }
        } else if(strlen(array_key_first($questions[$step]['options']))>1 && (strpos($data['chatid'], "telegram_") === 0 || strpos($data['chatid'], "webchat_") === 0 || strpos($data['chatid'], "wuzapi_") === 0)) {
            if(strpos($data['chatid'], "webchat_") === 0) {
                $max=99;
            } else {
                $max=2;
            }
            $unicodeChar = '\u0008';
            $null = json_decode('"'.$unicodeChar.'"');
            $buttons = array();
            $i=0;
            foreach($questions[$step]['options'] as $key=>$goto) {
                if($i>$max) break;
                $buttons[] = array("type"=>"reply","reply"=>array("id"=>"id".$i,"title"=>$key));
                $i++;
            }

            $body = array('body'=>array("text"=>$questions[$step]['body']),'action'=>array('buttons'=>$buttons),"type"=>"button");
            $question_body=$null.json_encode($body);
        }

        $questions[$step]['next_action']='same';
    }

    // Define new state for returning 
    $next = $questions[$step]['next_action'];
    if($next=='next') {
        $next_step = $step+1;
    } else if($next=='same') {
        $next_step = $step;
    } else {
        $next_step = intval($next);
    }

    // State construct
    $partialstate = array("previous_step"=>$step,"current_step"=>$next_step,"data"=>$botdata,"interaction_counter"=>$interaction_counter);
    if($wait>0) {
        $partialstate['wait']=$wait;
    }
    $state = json_encode($partialstate);

    return array($question_body,$state);
}

function validate_answer($n,$answer) {

    global $questions, $config, $botdata;

    if(!isset($botdata['answers'])) {
        $botdata['answers']=array();
    }

    if(!isset($questions[$n])) {
        return "[".$config['last_action']."]";
    }

    $answer = extractBody($answer);

    // check break out menu option
    if($config['breakout_word']!='') {
        if(trim(strtolower($answer))==trim(strtolower($config['breakout_word']))) {
            log_debug("break out word detected, go to ".$config['breakout_step']);
            return "[next=".$config['breakout_step']."]";
        }
    }

    if($questions[$n]['response_type']=='text') {
        $botdata['answers'][''.$n]=$answer;

        if(isset($questions[$n]['validation'])) {
            if($questions[$n]['validation']!='') {
                if(!preg_match("/".$questions[$n]['validation']."/",$answer)) {
                    return "[wrong]";
                }
            }
        }
        if($questions[$n]['next_action']!='next') { 
            $next_step = $questions[$n]['next_action'];
            return "[next=$next_step]";
        }
    } else 
    if($questions[$n]['response_type']=='range') {
        $valid_answers = $questions[$n]['options'];
        if(!in_array($answer,$valid_answers)) {
            return "[wrong]";
        } else {
            $botdata['answers'][''.$n]=$answer;
            if($questions[$n]['next_action']!='next') { 
                $next_step = $questions[$n]['next_action'];
                return "[next=$next_step]";
            }
        }
    } else
    if($questions[$n]['response_type']=='menu') {
        log_debug("validate menu, answer given: $answer");
        $valid_answers=array();
        foreach($questions[$n]['options'] as $key=>$val) {
            $valid_answers[strtolower($key)]=$val;
        }
        $found=0;
        foreach($valid_answers as $key=>$goto) {
            if (preg_match("/\b$key\b/i", $answer)) {
                 $found=1;
                 $next_step = $goto;
                 $botdata['answers'][''.$n]=$answer;
                 break;
            }
        }
        if($found==0) {
            return "[wrong]";
        }
        return "[next=$next_step]";
    }
    return '';
}

function send_request($method, $url, $body='', $hdrs, $auth=array()) {

    $headers = array();

    $options = array(
        CURLOPT_RETURNTRANSFER => 1,
        CURLOPT_CONNECTTIMEOUT => 10,
        CURLOPT_TIMEOUT        => 10,
        CURLOPT_HEADER         => 0,
        CURLOPT_SSL_VERIFYHOST => 0,
        CURLOPT_SSL_VERIFYPEER => 0
    );

    if($body!='') {
        $body = trim(preg_replace('/\s+/', '&', $body));
    }

    if($method=='POSTJSON') {

        $headers[] = 'Content-Type: application/json';

        if($body!='') {
            parse_str($body,$postfields);
            $payload = json_encode($postfields);
            $options[CURLOPT_POSTFIELDS] = $payload;
        } else {
            $options[CURLOPT_POSTFIELDS] = "{}";
        }
        $options[CURLOPT_POST] = 1;

    } else if($method=='POST') {

        if($body!='') {
            parse_str($body,$postfields);
            $options[CURLOPT_POSTFIELDS] = http_build_query($postfields);
            $options[CURLOPT_POST] = 1;
        }
    } else if(preg_match("/GET/",$method)) {
        if($body!='') {
            if(!preg_match("/\?/",$url)) { $url.="?"; } else { $url.="&"; }
            $url.=$body;
        }
    } else {
        // PUT, DELETE
        if($body!='') {
            parse_str($body,$postfields);
            $options[CURLOPT_POSTFIELDS] = http_build_query($postfields);
        }
        $options[CURLOPT_CUSTOMREQUEST] = $method;
    }

    log_debug("send request to $url method $method\n");

    foreach($hdrs as $passed_header) {
        $headers[]=$passed_header;
    }

    $options[CURLOPT_HTTPHEADER] = $headers;
    $options[CURLOPT_URL] = $url;

    if(isset($auth['username'])) {
        $options[CURLOPT_USERPWD] = $auth['username'] . ":" . $auth['password'];
    }

    $curl = curl_init();
    curl_setopt_array($curl, $options);
    $resp = curl_exec($curl);

    if($resp === false) { $ret = array('CURL_ERROR'=>curl_error($curl)); curl_close($curl); return $ret; }

    $httpcode = curl_getinfo($curl, CURLINFO_HTTP_CODE);

    curl_close($curl);

    if($method=='POSTJSON') {

        $ret = json_decode($resp,1);
        if($ret==null) {
            $ret = array('JSON_DECODE_ERROR'=>json_last_error_msg());
        }
        $ret['HTTP_CODE']=$httpcode;
        return $ret;
    } else {
        $ret = json_decode($resp,1);
        if($ret!=null) {
            // Returns array of decoded json response
            $ret['HTTP_CODE']=$httpcode;
            return $ret;
        } else {
            // Returns exact string returned from API call
            $final = strip_tags($resp);
            $response = array('HTTP_CODE'=>$httpcode,"Response"=>$final);
            return $response;
        }
    }
}

function expand_variables($body) {

    global $botdata;

    $dotFlatten = static function(array $item, $context = '') use (&$dotFlatten){
        $retval = [];
        foreach($item as $key => $value){
            $key = str_replace('$',"_",$key);
            $key = str_replace(':',"_",$key);
            if (\is_array($value) === true){
                $retval[$key.'_COUNT']=count($value);
                foreach($dotFlatten($value, "$context$key.") as $iKey => $iValue){
                    $retval[$iKey] = $iValue;
                }
            } else {
                $retval["$context$key"] = $value;
            }
        }
        return $retval;
    };

    if(is_array($botdata)) {
        foreach($dotFlatten($botdata) as $key=>$val) {
            $body = preg_replace("/{".$key."}/",$val,$body);
        }
    }

    return $body;
}

function downloadFile($url) {
    $tempFile = tempnam(sys_get_temp_dir(), 'download_');
    $fileHandler = fopen($tempFile, 'w');
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_FILE, $fileHandler);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    $success = curl_exec($ch);

    if (!$success) {
        fclose($fileHandler);
        unlink($tempFile);
        curl_close($ch);
        return array();
    }

    $pathInfo = pathinfo($url);
    $filename = $pathInfo['basename'];
    $mimeType = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
    curl_close($ch);
    fclose($fileHandler);

    // Return an associative array with the path, MIME type, and filename
    return array(
        'file' => $tempFile,
        'mime' => $mimeType,
        'filename' => $filename
    );
}

function log_debug($text) {
    global $DEBUG;
    $logfile = basename(__FILE__,'.php');
    if($DEBUG==1) {
        file_put_contents("/tmp/${logfile}.log",$text."\n",FILE_APPEND);
    }
}

function extractBody($msg) {
    $first_char = mb_substr($msg, 0, 1);
    if(mb_ord($first_char,'UTF-8')==8) {
        $msg = mb_substr($msg,1);
        $jsonObj = json_decode($msg,1);
        $msg = isset($jsonObj['body'])?$jsonObj['body']:$msg;
    }
    return $msg;
}
